import logging
import os


def get_root_dir() -> str:
    """
    This function retrieves the root dir of this package.
    """
    return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def get_logger(log_file: str = "logs.log") -> logging.Logger:
    """
    This function creates a logger that we use through out the package.

    Args:
        log_file: relative file name for the log file to be stored in src/utils/../../logs/.
    """
    root_dir = get_root_dir()
    log_dir = os.path.join(root_dir, "logs")
    if "logs" not in os.listdir(root_dir):
        os.mkdir(log_dir)
    if log_file == "logs.log" and log_file in os.listdir(log_dir):
        os.remove(os.path.join(log_dir, log_file))
    if log_file in os.listdir(log_dir):
        raise ValueError(f"the provided log file already exists : {log_file}")
    formatter = logging.Formatter("%(message)s")
    handler_info = logging.FileHandler(
        os.path.join(log_dir, log_file), mode="a", encoding="utf-8"
    )
    handler_info.setFormatter(formatter)
    handler_info.setLevel(logging.INFO)
    logger = logging.getLogger("upperbound_logger")
    logger.setLevel(logging.INFO)
    logger.addHandler(handler_info)
    return logger



from typing import Any, Optional
import importlib.util
import sys
import os
import logging
from importlib.metadata import version

PACKAGES_HAVE_BEEN_CHECKED = False


def check_python(logger: Optional[logging.Logger] = None) -> None:
    """
    This function checks that the python version in use is 3+.

    Args:
        logger: logs the package check
    """
    if sys.version_info[0] == 3:
        if sys.version_info[1] >= 9:
            print(f"\033[92m\u2714\033[0m Python version ", end="")
            print(f"{sys.version_info[0]}.{sys.version_info[1]}", end="")
            print(f".{sys.version_info[2]}")
            if logger is not None:
                logger.info(
                    f"\033[92m\u2714\033[0m Python version "
                    f"{sys.version_info[0]}.{sys.version_info[1]}"
                    f".{sys.version_info[2]}"
                )
        else:
            print(
                f"you are using python{sys.version_info[0]}" f".{sys.version_info[1]}"
            )
            print(f"this code was tested with python 3.9.7")
            print(f"you might need to upgrade your version")
            if logger is not None:
                logger.info(
                    f"you are using python{sys.version_info[0]}"
                    f".{sys.version_info[1]}"
                )
                logger.info(f"this code was tested with python 3.9.7")
                logger.info(f"you might need to upgrade your version")
    else:
        print(f"you are using python{sys.version_info[0]}")
        print(f"this code only works in python 3")
        if logger is not None:
            logger.info(f"you are using python{sys.version_info[0]}")
            logger.info(f"this code only works in python 3")


def check_if_a_module_exists(module_name: str) -> bool:
    """
    This function checks if a module can be loaded
    """
    if module_name in sys.modules:
        return True
    elif (importlib.util.find_spec(module_name)) is not None:
        return True
    else:
        return False


def check_module_version(module_name: str) -> Any:
    """
    This function gets the version of a module

    Args:
        module_name: name of the python module to check
    """
    if check_if_a_module_exists(module_name=module_name):
        if module_name == "batch_normalization_folding":
            return version("tensorflow-batchnorm-folding")
        module = importlib.import_module(module_name)
        if module_name == "dotenv":
            return "0.20.0"
        elif module_name == "nvidia_smi":
            return "7.352.0"
        elif module_name == "official":
            return "2.9.2"
        return module.__version__
    return None


def create_model_folder() -> None:
    """
    This functions creates a folder where we can save the quantized models
    """
    if "quantized_models" not in os.listdir(".."):
        os.mkdir("../quantized_models")


def compare_versions(v1, v2) -> bool:
    """
    This function comapres two versions, usual use consist in
    ensuring that a module has a satisfactory version for stable use

    Args:
        module_name: name of the python module to check
    """
    if v1 is None:
        return False
    if v1[0] > v2[0]:
        return True
    elif v1[0] == v2[0]:
        if v1[1] > v2[1]:
            return True
        elif v1[1] == v2[1]:
            if v1[2] >= v2[2]:
                return True
    return False


def check_packages(logger: Optional[logging.Logger] = None) -> None:
    """
    checks that all the required packages are installed

    Args:
        logger: logs the package check
    """
    global PACKAGES_HAVE_BEEN_CHECKED
    if not PACKAGES_HAVE_BEEN_CHECKED:
        print()

        check_python(logger=logger)
        module_v_int = []
        module_v_str = []
        pip_commands = [
            "pip install --upgrade tensorflow",
            "pip install --upgrade scikit-image",
            "pip install --upgrade opencv-python",
            "pip install --upgrade Pillow",
        ]
        missing_modules = False
        packages_to_check = ["tensorflow", "skimage", "cv2", "PIL"]
        recommended_version = [(2, 10, 0), (0, 19, 3), (4, 6, 0), (9, 2, 0)]
        for package in packages_to_check:
            module_version = check_module_version(module_name=package)
            if module_version is None:
                module_v_int.append(None)
                module_v_str.append("-")
                missing_modules = True
            else:
                module_v_str.append(module_version)
                module_v_int.append([int(e) for e in module_version.split(".")])

        pip_commands_to_run = []
        column_0 = ["package"]
        column_1 = ["current version"]
        column_2 = ["recommended version"]
        column_3 = ["checks"]
        for (version, v_str, reco_version, pip_command, name) in zip(
            module_v_int,
            module_v_str,
            recommended_version,
            pip_commands,
            packages_to_check,
        ):
            if compare_versions(version, reco_version):
                column_3.append("\u2714")
            else:
                column_3.append("\u2717")
                pip_commands_to_run.append(pip_command)
            column_2.append(f"{reco_version[0]}.{reco_version[1]}.{reco_version[2]}")
            column_1.append(v_str)
            column_0.append(name)

        l0 = len(max(column_0, key=len))
        l1 = len(max(column_1, key=len))
        l2 = len(max(column_2, key=len))
        l3 = len(max(column_3, key=len))
        limiter = "+" + (l0 + l1 + l2 + l3 + 11) * "-" + "+"
        print(limiter)
        if logger is not None:
            logger.info(limiter)
        for cpt, (s0, s1, s2, s3) in enumerate(
            zip(column_0, column_1, column_2, column_3)
        ):
            if cpt == 0:
                color = "\033[1m"
                end_c = "\033[0m"
                color_ = ""
                end_c_ = ""
                assert_color = ""
                assert_end_c = ""
            else:
                color = ""
                end_c = ""
                color_ = "\033[4m"
                end_c_ = "\033[0m"
                assert_color = "\033[92m"
                if s3 == "\u2717":
                    assert_color = "\033[91m"
                assert_end_c = "\033[0m"
            print(
                f"| {color}{s0.center(l0)}{end_c} | {color}{s1.center(l1)}{end_c} "
                f"| {color}{(color_ + s2 + end_c_).center(l2 + len(end_c_) + len(color_))}{end_c} "
                f"| {assert_color}{color}{s3.center(l3)}{end_c}{assert_end_c} |"
            )
            if logger is not None:
                logger.info(
                    f"| {color}{s0.center(l0)}{end_c} | {color}{s1.center(l1)}{end_c} "
                    f"| {color}{(color_ + s2 + end_c_).center(l2 + len(end_c_) + len(color_))}{end_c} "
                    f"| {assert_color}{color}{s3.center(l3)}{end_c}{assert_end_c} |"
                )
        print(limiter)
        if logger is not None:
            logger.info(limiter)

        PACKAGES_HAVE_BEEN_CHECKED = True
        if len(pip_commands_to_run) != 0:
            print("to install/upgrade missing dependencies please use:")
            if logger is not None:
                logger.info("to install/upgrade missing dependencies please use:\n")
            for pip in pip_commands_to_run:
                print(f"\t{pip}")
                if logger is not None:
                    logger.info(f"\t{pip}")
        if missing_modules:
            sys.exit()


if __name__ == "__main__":
    check_packages()



from .grid_sampler import bilinear_sampler
import tensorflow as tf
from skimage import transform as trans
import numpy as np
from typing import Tuple, Optional


def atanh(x: tf.Tensor) -> tf.Tensor:
    """
    Implements the atanh function

    Args:
        x: tensor to activate element-wise.
    """
    return 0.5 * tf.math.log((1 + x) / (1 - x))


def apply_mat_tensor(coords: np.ndarray, matrix: np.ndarray) -> np.ndarray:
    """
    Applies the matrix tensor to the coordinates

    Args:
        coords: array of coordinates to modify
        matrix: matrix array for coords change
    """
    matrix.astype(np.float32)
    coords = np.array(coords, copy=False, ndmin=2)
    perm = np.arange(len(coords.shape))
    perm[0] = 1
    perm[1] = 0
    x, y = np.transpose(coords, axes=perm)
    src = np.stack([x, y, np.ones_like(x)], axis=0)
    dst = src.T @ matrix.T
    dst[dst[:, 2] == 0, 2] = np.finfo(float).eps
    dst[:, :2] /= dst[:, 2:3]
    return dst[:, :2]


def boxes2landmark(boxes):
    """
    converts boxes to landmarks from which we will derive the similarity transform

    Args:
        boxes: bounding box of the face
    """
    landmarks = []
    for i in range(boxes.shape[0] - 1):
        cen_x = (boxes[i][0] + boxes[i][2]) / 2.0
        cen_y = (boxes[i][1] + boxes[i][3]) / 2.0
        landmarks.append((cen_x, cen_y))
    mouth1 = (boxes[3][0], boxes[3][3])
    landmarks.append(mouth1)
    mouth2 = (boxes[3][2], boxes[3][3])
    landmarks.append(mouth2)
    return np.array(landmarks)


def get_warped_coords(
    corrds: np.ndarray,
    tform: trans.SimilarityTransform,
    tform2: trans.SimilarityTransform,
) -> tf.Tensor:
    """
    computes the warped coordinates using tanh roi

    Args:
        corrds: initial coordinates
        tform: first transformation to apply
        tform2: second transformation to apply
    """
    matrix1 = np.linalg.inv(tform.params)
    matrix2 = tform2.params
    grid = apply_mat_tensor(
        atanh(tf.clip_by_value(apply_mat_tensor(corrds, matrix2), -0.9999, 0.9999)),
        matrix1,
    )
    return grid


def get_inversed_coords(
    corrds: np.ndarray,
    tform: trans.SimilarityTransform,
    tform2: trans.SimilarityTransform,
) -> tf.Tensor:
    """
    computes the warped coordinates using inverted tanh roi

    Args:
        corrds: initial coordinates
        tform: first transformation to apply
        tform2: second transformation to apply
    """
    matrix1 = tform.params
    matrix2 = np.linalg.inv(tform2.params)
    grid = apply_mat_tensor(tf.math.tanh(apply_mat_tensor(corrds, matrix1)), matrix2)
    return grid


def get_coords(
    tform: trans.SimilarityTransform,
    tform2: trans.SimilarityTransform,
    out_shape: Tuple[int],
    mode: str = "warp",
) -> tf.Tensor:
    """
    extract coords

    Args:
        out_shape: shape of the output
        tform: first transformation to apply
        tform2: second transformation to apply
        mode: mode to use for the method
    """
    cols, rows = out_shape[0], out_shape[1]
    tf_coords = np.indices((cols, rows), dtype=np.float32).reshape(2, -1).T
    if mode == "warp":
        tf_coords = get_warped_coords(
            tf_coords,
            tform=tform,
            tform2=tform2,
        )
    elif mode == "inverse":
        tf_coords = get_inversed_coords(
            tf_coords,
            tform=tform,
            tform2=tform2,
        )
    tf_coords = tf.transpose(
        tf.reshape(tf.transpose(tf_coords), (-1, cols, rows)), perm=(0, 2, 1)
    )
    return tf_coords


def coords2grid(coords: tf.Tensor, in_image_shape: Tuple[int]) -> tf.Tensor:
    """
    converts coords to grip for grid sampling

    Args:
        coords: coordiantes to convert
        in_image_shape: image shape
    """
    ih, iw = in_image_shape[0], in_image_shape[1]
    coords = (2 * coords) / tf.expand_dims(
        tf.expand_dims((tf.stack([np.float32(ih) - 1, np.float32(iw) - 1])), axis=-1),
        axis=-1,
    ) - 1
    grid = tf.expand_dims(tf.transpose(coords, (1, 2, 0)), axis=0)
    return grid


def FastTanhWarping(
    image: tf.Tensor, boxes: np.ndarray, output_size: Optional[Tuple[int]] = None
):
    """
    applies fast tanh warping using skimage transforms

    Args:
        boxes: bounding box of the face
        output_size: size of the output iamge
    """
    tform = trans.SimilarityTransform()
    dst = np.array([[-0.25, -0.1], [0.25, -0.1], [0.0, 0.1], [-0.15, 0.4], [0.15, 0.4]])
    tform2 = trans.SimilarityTransform(
        scale=1.0 / 256.0, rotation=0, translation=(-1, -1)
    )
    landmarks = boxes2landmark(boxes=boxes)
    tform.estimate(landmarks, dst)
    if output_size is None:
        output_size = np.array(image).shape
    corrds = get_coords(tform, tform2, output_size, mode="warp")
    grid = coords2grid(corrds, np.array(image).shape)
    warped_image = bilinear_sampler(
        img=np.expand_dims(image, axis=0), x=grid[..., 0], y=grid[..., 1]
    )
    return warped_image



import os
import shutil


def remove_chache_folders(current_repo: str = ""):
    """
    remove __pycache__ files recursively
    
    Args:
        current_repo: current folder to purge
    """
    if current_repo == "":
        new_refs = [elem for elem in os.listdir()]
    else:
        new_refs = [current_repo + "/" + elem for elem in os.listdir(current_repo)]
    for elem in new_refs:
        if os.path.isdir(elem):
            if "__pycache__" in elem:
                shutil.rmtree(elem)
            else:
                remove_chache_folders(current_repo=elem)



from .fastwarping import FastTanhWarping
from typing import Tuple, List
import tensorflow as tf
import numpy as np
from PIL import Image


def labels2boxes(
    input: np.ndarray,
    left_eye_class: List[int] = [2, 4],
    right_eye_class: List[int] = [3, 5],
    nose_class: List[int] = [6],
    mouth_class: List[int] = [7, 8, 9],
) -> tf.Tensor:
    """
    extracts bbox from labels

    Args:
        input: input labels
        left_eye_class: class indices associated with the left eye
        right_eye_class: class indices associated with the right eye
        nose_class: class indices associated with the nose
        mouth_class: class indices associated with the mouth
    """
    images = np.transpose(input, axes=(2, 0, 1))
    leye_image = np.sum(images[left_eye_class], axis=0)
    reye_image = np.sum(images[right_eye_class], axis=0)
    nose_image = np.sum(images[nose_class], axis=0)
    mouth_image = np.sum(images[mouth_class], axis=0)
    leye = Image.fromarray(leye_image).getbbox()
    reye = Image.fromarray(reye_image).getbbox()
    nose = Image.fromarray(nose_image).getbbox()
    mouth = Image.fromarray(mouth_image).getbbox()
    if leye is None:
        leye = [0, 0, 0, 0]
    if reye is None:
        reye = [0, 0, 0, 0]
    if nose is None:
        nose = [0, 0, 0, 0]
    if mouth is None:
        mouth = [0, 0, 0, 0]
    boxes = np.array((leye, reye, nose, mouth))
    assert boxes.shape == (4, 4)
    return boxes


def sharpen_labels(label: tf.Tensor) -> tf.Tensor:
    """
    ensures that the labels are ones an zeros

    Args:
        label: one hot labels (bilinear)
    """
    return tf.math.round(tf.nn.softmax(logits=label, axis=-1))


def TanhWarping(
    image: np.ndarray, label: np.ndarray, size: Tuple[int]
) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    apply tanh warping based on
    https://openaccess.thecvf.com/content_CVPR_2019/papers/Lin_Face_Parsing_With_RoI_Tanh-Warping_CVPR_2019_paper.pdf

    Args:
        image: images to rectify
        label: one hot labels
        size: output target size
    """
    boxes = labels2boxes(np.array(label))
    warped_image = FastTanhWarping(image=image, boxes=boxes, output_size=size)
    warped_labels = FastTanhWarping(image=label, boxes=boxes, output_size=size)
    warped_labels = sharpen_labels(warped_labels)
    return warped_image, warped_labels



import tensorflow as tf
import numpy as np
from typing import Tuple
import cv2
import sys
import os


class labelConvertor_helen:
    def __init__(self):
        super(labelConvertor_helen, self).__init__()
        self.correspondance = {
            0: "unknown",
            1: "unknown",
            2: "unknown",
            3: "unknown",
            4: "unknown",
            5: "unknown",
            6: "unknown",
            7: "unknown",
            8: "unknown",
            9: "unknown",
            10: "unknown",
            11: "unknown",
        }

    def __call__(self, map: tf.Tensor) -> tf.Tensor:
        map[np.where(map >= 11)] = -1

        return tf.one_hot(map, depth=len(self.correspondance) - 1).numpy()


mask_convertor = labelConvertor_helen()


def decode_image(image_path: str, ismask: bool) -> np.ndarray:
    """
    load and converts an iamge or label

    Args:
        image_path: path to file
        ismask: boolean to know if we are dealing with a label or not
    """
    if ismask:
        decoded_image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
    else:
        decoded_image = cv2.imread(image_path)
    if decoded_image is None:
        print(image_path, "error")
        sys.exit()
    if ismask:
        global mask_convertor
        image = decoded_image.astype(np.uint8)
        image[np.where(image >= 11)] = -1
        image = mask_convertor(image)
    else:
        image = decoded_image.astype(np.float32)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB).astype(np.float32)
    return image


def run_a_unit_test() -> Tuple[np.ndarray, np.ndarray]:
    """
    collects an example for unit testing
    """
    return (
        decode_image(
            image_path=os.path.join(os.getcwd(), "unit_test", "2546515404_1_image.jpg"),
            ismask=False,
        ),
        decode_image(
            image_path=os.path.join(os.getcwd(), "unit_test", "2546515404_1_label.png"),
            ismask=True,
        ),
    )


def save_example_pair(new_image: np.ndarray, new_label: np.ndarray) -> None:
    """
    save the warped images and corresponding labels

    Args:
        new_image: image to save
        new_label: warped label to save
    """
    if not isinstance(new_image, np.ndarray):
        new_image = new_image.numpy()
    if not isinstance(new_label, np.ndarray):
        new_label = new_label.numpy()
    cv2.imwrite(
        os.path.join(os.getcwd(), "unit_test", "new_image.jpg"),
        cv2.cvtColor(new_image, cv2.COLOR_BGR2RGB),
    )



from typing import Tuple
import os

if __name__ == "__main__":
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

    import argparse

    parser = argparse.ArgumentParser(
        usage=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "--logger",
        nargs="?",
        type=str,
        default="logs.log",
        help="path to save the logs",
    )
    parser.add_argument(
        "--unit_test",
        action="store_true",
        default=False,
        help="run the unit test",
    )
    args = parser.parse_args()
    from .utils.getlogger import get_logger

    logger = get_logger(log_file=args.logger)
else:
    logger = None
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from .utils.packages import check_packages
from .warping.warping import TanhWarping

check_packages(logger=logger)
import tensorflow as tf
import numpy as np


def apply_roi_tanh_warping(
    image: np.ndarray, label: np.ndarray, size: Tuple[int]
) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    applies a warping to an example pair

    Args:
        image: images to rectify
        label: one hot labels
        size: output target size
    """
    return TanhWarping(image=image, label=label, size=size)


if __name__ == "__main__":
    if args.unit_test:
        from .utils.test import run_a_unit_test, save_example_pair

        image, label = run_a_unit_test()
        new_image, new_label = apply_roi_tanh_warping(
            image=image, label=label, size=(622, 639)
        )
        save_example_pair(new_image=new_image, new_label=new_label)
    from .utils.clear import remove_chache_folders

    remove_chache_folders()



"""
from https://github.com/kevinzakka/spatial-transformer-network/blob/master/stn/transformer.py#L159
"""
import tensorflow as tf


def get_pixel_value(img: tf.Tensor, x: tf.Tensor, y: tf.Tensor) -> tf.Tensor:
    """
    Utility function to get pixel value for coordinate
    vectors x and y from a  4D tensor image.
    Input
    -----
    - img: tensor of shape (B, H, W, C)
    - x: flattened tensor of shape (B*H*W,)
    - y: flattened tensor of shape (B*H*W,)
    Returns
    -------
    - output: tensor of shape (B, H, W, C)
    """
    shape = tf.shape(x)
    batch_size = shape[0]
    height = shape[1]
    width = shape[2]

    batch_idx = tf.range(0, batch_size)
    batch_idx = tf.reshape(batch_idx, (batch_size, 1, 1))
    b = tf.tile(batch_idx, (1, height, width))

    indices = tf.stack([b, y, x], 3)

    return tf.gather_nd(img, indices)


def bilinear_sampler(img: tf.Tensor, x: tf.Tensor, y: tf.Tensor) -> tf.Tensor:
    """
    Performs bilinear sampling of the input images according to the
    normalized coordinates provided by the sampling grid. Note that
    the sampling is done identically for each channel of the input.
    To test if the function works properly, output image should be
    identical to input image when theta is initialized to identity
    transform.
    Input
    -----
    - img: batch of images in (B, H, W, C) layout.
    - grid: x, y which is the output of affine_grid_generator.
    Returns
    -------
    - out: interpolated images according to grids. Same size as grid.
    """
    H = tf.shape(img)[1]
    W = tf.shape(img)[2]
    max_y = tf.cast(H - 1, "int32")
    max_x = tf.cast(W - 1, "int32")
    zero = tf.zeros([], dtype="int32")

    # rescale x and y to [0, W-1/H-1]
    x = tf.cast(x, "float32")
    y = tf.cast(y, "float32")
    x = 0.5 * ((x + 1.0) * tf.cast(max_x - 1, "float32"))
    y = 0.5 * ((y + 1.0) * tf.cast(max_y - 1, "float32"))

    # grab 4 nearest corner points for each (x_i, y_i)
    x0 = tf.cast(tf.floor(x), "int32")
    x1 = x0 + 1
    y0 = tf.cast(tf.floor(y), "int32")
    y1 = y0 + 1

    # clip to range [0, H-1/W-1] to not violate img boundaries
    x0 = tf.clip_by_value(x0, zero, max_x)
    x1 = tf.clip_by_value(x1, zero, max_x)
    y0 = tf.clip_by_value(y0, zero, max_y)
    y1 = tf.clip_by_value(y1, zero, max_y)

    # get pixel value at corner coords
    Ia = get_pixel_value(img, x0, y0)
    Ib = get_pixel_value(img, x0, y1)
    Ic = get_pixel_value(img, x1, y0)
    Id = get_pixel_value(img, x1, y1)

    # recast as float for delta calculation
    x0 = tf.cast(x0, "float32")
    x1 = tf.cast(x1, "float32")
    y0 = tf.cast(y0, "float32")
    y1 = tf.cast(y1, "float32")
    # calculate deltas
    wa = (x1 - x) * (y1 - y)
    wb = (x1 - x) * (y - y0)
    wc = (x - x0) * (y1 - y)
    wd = (x - x0) * (y - y0)

    # add dimension for addition
    wa = tf.expand_dims(wa, axis=3)
    wb = tf.expand_dims(wb, axis=3)
    wc = tf.expand_dims(wc, axis=3)
    wd = tf.expand_dims(wd, axis=3)

    # compute output
    out = tf.add_n([wa * Ia, wb * Ib, wc * Ic, wd * Id])

    return tf.squeeze(out)
